/** @file  VLBI_Frame.hpp 
 *  A interface Class for VLBI Frame.
 *   This is succceded by K5/VSSP32, Mark5B, and VDIF frames
 * @author M.Sekido
 * @Date 2009.4.4
 */


#ifndef __VLBI_Frame_HPP__
#define __VLBI_Frame_HPP__


#include "DAS_info.hpp"
#include "Mk5B_Header.hpp"
#include "K5VSSP_Header.hpp"
#include "VDIF_Header.hpp"
#include "Trans_info.hpp"
#include "Logsystem.hpp"
//#include "k5vsi_mdl.hpp"
#include "crc.hpp"

extern "C"{
#include <time.h>
}

/// Data Acquisition System Types
//enum DAS_t {K5VSSP, ADS2000, ADS1000, ADS3000, ADS3000plus, MK5B, Undef};

/** 
 * Frame size: 
 * VDIF Frame size must be multiple of 8bytes.
 * - IP header size = 24 bytes;
 * - UDP header size = 8 bytes
 * - SUDP  header size = 8byte
 * - VDIF header size = 32 bytes;
 * - Available space for application = 1500-32 = 1468
 * - Application Transport Protocol Header (ATPH) = 8 bytes
 * - Multiple of 8byte for VLBI_frame = 1456 at maximum.
 */

const int DEFAULT_VDIF_FRAMESIZE = 1456; 

/// Function converting tm data to mjd
double tm2mjd(struct tm date);
/// Function for Computing MJD from year/moth/day
double cal2mjd(int day, int month, int year);

/** @class VLBI_Frame 
 * A Abstract interface class for general VLBI data-frames.
 *  VLBI_Frame as abstract interface class for general VLBI data-frames.
 */

class VLBI_Frame{  
public:
  /// VLBI Frame Types 
  enum VLBI_Frame_t {VDIF_frm, MK5B_frm, K5VSSP_frm, K5VSI_frm, Undef_frm};
protected:
    int32_t framesize;     ///< Frame (data part) size (byte)
    //  int headsize;      ///< Header size (Bytes)
    DAS_Info das_inf;      ///< Sampling parameters  
    Time_dat data_epoch;   ///< Data Epoch 
    uint16_t  iframe;      ///< Number of Frame
public:
  /** Constructor */
  VLBI_Frame():framesize(0),iframe(0) {};
  /** Constructor */
  VLBI_Frame(const int fs):framesize(fs),iframe(0){};
  //  VLBI_Frame(const VLBI_Frame_t ft=VDIF_frm){Type_Set(ft);}
  /// Abstract function of Destructor 
  virtual ~VLBI_Frame(){ }; ///< Destructor 
  //  void Set_DAS_Info(DAS_Info& d){das_info = d;}
  //  int Type_Set(const VLBI_Frame_t ft); //< VLBI Frame type setting
  /** Set frame header and payload data (size byte) from data src */
  virtual int Make_frame( uint8_t* pkt, const uint8_t* src, const int size )=0;
  virtual const VLBI_Frame_t Type() const=0; ///< Return VLBI Frame type
  virtual string TypeName()=0; ///< Returns Frame type in string 
  /// Framesize (bytes) including header 
  int32_t Framesize()const{return framesize;}
  /** Set Sampling parameters, return data rate (Bytes/sec)  */
  virtual int32_t Set_Sparam(const int n, const int q, const uint32_t Hz);
  /// Abstract function to Return Headsize (byte) of the frame-type
  virtual int Headsize()const=0; // {return headsize;}
  //  void Set_Proto(VLBI_Proto* p){proto=p;}
  /// Set DAS information
  virtual const int Set_DasInfo(const DAS_Info& ds){das_inf = ds; return 0;}
  /// Returns reference of das info.
  const DAS_Info& DAS_Inf()const {return das_inf;}
  /// Abstract Function to set Epoch to the header 
  virtual void Set_Epoch(const Time_dat& t1){data_epoch=t1;}
  /// Abstract Function to set Epoch to the header 
  virtual void Set_Epoch(const time_t t1);
  /// Returns Epoch in Time_dat class.
  Time_dat Get_Epoch()const{return data_epoch;}
  virtual void Set_Epoch0(const time_t t1){;} ///< Do nothing except for VDIF
  /// Set Header and frame number to the packet 
  virtual int Set_header(const uint8_t* buf, const int i=-1){return 0;}
  /// Reset Frame counter number 
  virtual void Reset_header(){;}
  /// Set Framesize 
  virtual void Set_Framesize(const uint32_t sz)=0;
  /// Set Headr and DAS parameter to the packet.
  virtual void Set_Head_par(const time_t t1, DAS_Info* dasinf=0){
    Set_Epoch(t1); 
    if(dasinf != 0) das_inf = *dasinf; 
  }
    //  int Check();
};

inline  void VLBI_Frame::Set_Epoch(const time_t t1){
    struct tm tm1;
    time_t t2 = t1;
    gmtime_r(&t2, &tm1);
    int yy = tm1.tm_year + 1900;
    int mon = tm1.tm_mon + 1;
    int day = tm1.tm_mday;
    int h = tm1.tm_hour;
    int m = tm1.tm_min;
    int s = tm1.tm_sec;
    data_epoch.Set_time(yy,mon,day,h,m,s);
}



//========== VDIF ===from Here  ====================
/** VDIF Frame class.
 * Secceeding from VLBI_Frame class 
 */
class VDIF_Frame: public VLBI_Frame{
  //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
  VDIF_Hdr vdif_hdr;   ///< VDIF Header prototype  
  //  vector <uint8_t> packet; //< packet Frame including header
  //  time_t ref_epoch0; ///< Reference Epoch 0
  Time_dat ref_epoch0; ///< Reference Epoch 
public:
    //  VDIF_Frame(): k5vsi_dev(0){}
    //  VDIF_Frame():ref_epoch0(0){;}
  VDIF_Frame(); ///< Constructor 
  /// Returns VLBI-Frame type in number 
  const VLBI_Frame_t Type()const{return VDIF_frm;}
  /// Returns VLBI-Frame type in string
  string TypeName(){return string("VDIF");}
  /** Set Framesize (byte) */
  void Set_Framesize(const uint32_t sz){
    framesize =sz; vdif_hdr.Set_FrameLen(sz);}
  /** Setting frame size by number of divisions/sec. */
  int32_t Set_FrameDiv(const int n=10);
  //  int Decompose_frame(const uint8_t* buf, uint32_t size ){return 0;}
  /** Set frame header and payload data (size byte) from data src */
  int Make_frame( uint8_t* pkt, const uint8_t* src, const int size );
  /// Framesize (bytes) including header 
  //  uint32_t Framesize()const{return packet.size();};
  /// Returns Headersize (bytes)
  int Headsize()const{return VDIF_Hdr_size;}
  /// Set K5VSI  device module
  //  void Set_K5VSI( K5_VSI *k){k5vsi_dev = k;}
  /** Set Sampling parameters, s:(Hz),  return data rate (Bytes/sec)  */
  int32_t Set_Sparam(const int n, const int q, const uint32_t Hz);
  /// Set Initial Epoch
  void Set_Epoch0(const time_t t1){ ref_epoch0 = vdif_hdr.Set_Epoch0(t1);}
  /// Set Epoch to the header
  void Set_Epoch(const time_t t1);
  /// Set Epoch to the header
  void Set_Epoch(const Time_dat& td);
  /// Set Header to the packet 
  int  Set_header(const uint8_t* buf, const int ifrm=-1);
  /// Set Station ID 
  void Set_StnID(const char* str){vdif_hdr.Set_Stnid(str);}
  /// Reset sequential number of header
  void Reset_header(){vdif_hdr.Frame_Number=0;} 
  const int Set_DasInfo(const DAS_Info& ds);
  /// Set Headr and DAS parameter to the packet.
  void Set_Head_par(const time_t t1, DAS_Info* dasinf=0);
};

inline  const int VDIF_Frame::Set_DasInfo(const DAS_Info& ds){
  das_inf = ds;
  vdif_hdr.Set_Nch(ds.Nch());
  vdif_hdr.Set_Qbit(ds.Qbit());
  vdif_hdr.Set_Srate(ds.Srate()/1000);
  return 0;
}



inline  void VDIF_Frame::Set_Head_par(const time_t t1, DAS_Info* dasinf){
    VLBI_Frame::Set_Head_par(t1,dasinf);
    if(dasinf !=0){
	vdif_hdr.Set_Nch(dasinf->Nch());
	vdif_hdr.Set_Qbit(dasinf->Qbit());
	vdif_hdr.Set_Srate(dasinf->Srate()/1000); // input in kHz
    }
}

inline int VDIF_Frame::Make_frame( uint8_t* pkt, const uint8_t* src, 
				   const int tobe_pkt){
  Set_header(pkt);
  pkt += VDIF_Hdr_size;
  memcpy( pkt, src, tobe_pkt );
  return tobe_pkt + VDIF_Hdr_size;
}

inline  void VDIF_Frame::Set_Epoch(const time_t t1){
    VLBI_Frame::Set_Epoch(t1);
    vdif_hdr.Set_Epoch(data_epoch);
}

inline void VDIF_Frame::Set_Epoch(const Time_dat& t1){
    data_epoch=t1;
    vdif_hdr.Set_Epoch(data_epoch);
}



/** Set frame size by dividing pagesize(1sec) with a integer number */
inline int32_t VDIF_Frame::Set_FrameDiv(const int n){
    int32_t byte_sec = das_inf.Bytes_sec();
    int n1 =n;
    while( byte_sec % (8*n1) !=0 && byte_sec > 8*n1){
	n1++;
    }
    int32_t pld_size = byte_sec/n1;
    framesize = pld_size + VDIF_Hdr_size;
    vdif_hdr.Set_FrameLen(framesize);
    fprintf(stderr,
	    "Info: VDIF_Frame::Set_FrameDiv(%d) fixed Ndiv=%d, Framesize=%d\n",
	    n,n1,framesize);
    return framesize;
}

/** Set Sampling parameters, s:(Hz),  return data rate (Bytes/sec)  */
inline 
int32_t VDIF_Frame::Set_Sparam(const int n, const int q, const uint32_t Hz){
    vdif_hdr.Set_Sparam( n, q, Hz ); 
    das_inf.Set_Param( n, q, Hz );
    //  framesize=vdif_hdr.Framesize();
    return das_inf.Speed()*4;
}


inline int VDIF_Frame::Set_header(const uint8_t* buf, const int ifrm){

  if(ifrm >=0)  vdif_hdr.Frame_Number  = ifrm;  
  memcpy((void*)buf, &vdif_hdr, VDIF_Hdr_size);
  
  vdif_hdr.Frame_Number++; // Count up
  
  return VDIF_Hdr_size;
}

//========== VDIF ===to   Here  ====================

//========== Mk5B ===from Here  ====================
class Mk5B_Frame: public VLBI_Frame{
    //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
    Mk5B_Hdr m5hdr;  /// Mark5B header
    CRC_mdl*  pCRC_mdl;
    /// Frame counter 
public:
  /// Default Constructor 
  Mk5B_Frame():VLBI_Frame(DEFAULT_MK5B_FRAMESIZE){ pCRC_mdl = new CRC_jive;}
  ~Mk5B_Frame(){delete pCRC_mdl;} ///< Destructor 
  /// Return Frame type
  const VLBI_Frame_t Type()const{return MK5B_frm;}
  string TypeName(){return string("Mk5B");}
  /// Decomposing frame 
  //  int Decompose_frame(const uint8_t* buf, uint32_t size );
  /// Writing Frame 
  int Make_frame( uint8_t* pkt,  const uint8_t* src, const int data_size );
  /// Framesize (bytes) including header 
  //  uint32_t Framesize()const{return m5hdr.Framesize();};
  /// Headersize (bytes)
  int Headsize()const{return Mk5B_Hdr_size;}
  int Set_header(const uint8_t* buf, const int ifrm=-1);
  void Reset_header(){m5hdr.iframe=0;} ///< Reset sequential number of header
  int Convert();
  /// Set Mark5 header
  //   void Set_mark5header( uint32_t *header, 
  // 			const time_t epoch, const double sec, 
  // 			const uint16_t iblk);
  void Set_Framesize(const uint32_t sz){;}
};


inline int Mk5B_Frame::Make_frame( uint8_t* pkt,  const uint8_t* src,
				   const int data_size){
    Set_header(pkt);
    pkt += Mk5B_Hdr_size;
    memcpy( pkt, src, data_size );
    return data_size + Mk5B_Hdr_size;
}

//========== Mk5B ===to   Here  ====================
//========== K5VSSP ===from Here  ====================

class K5VSSP32_Frame: public VLBI_Frame{
    //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
    K5VSSP32_Hdr k5hdr;  ///<K5VSSP32 Hearder
    string stname;
    string stid;
public:
  /// Default constructor
  K5VSSP32_Frame():stname("Kashim34"),stid("kb"){;}
  /// Returns VLBI Frame type by number
  const VLBI_Frame_t Type()const{return K5VSSP_frm;}
  /// Returns VLBI Frame type by string
  string TypeName(){return string("K5VSSP32");}
  /** Set up frame copying header and data
   * @param buf: buffer of the packet
   * @param src: data buffer
   * @param data_size: size of data in bytes
   */
  int Make_frame( uint8_t* buf, const uint8_t* src, const int data_size );
  /// Returns Headersize (bytes)
  int Headsize()const{return K5VSSP32_Hdr_size;}
  /// Set Sampling parameters
  int Set_Sparam(const int n, const int q, const uint32_t s);
  void Set_Epoch(const Time_dat& t1); ///< Set Epoch
  void Set_Epoch(const time_t t1); ///< Set Epoch
  /** Set Header info to the packet 
   * @param buf: buffer of the packet
   * @param ifrm: Frame number. If negative, just count up. 
   */
  int Set_header(const uint8_t* buf, const int ifrm=-1);
  /// Set Station name and Station ID
  void Set_StnameID(const char* name, const char* id=0){
	stname = name; k5hdr.Set_Stname(stname.c_str());
	if(id!=0){ stid = id;k5hdr.Set_StnID(id);}}
  /// Returns Pointer of Header data 
  const K5VSSP32_Hdr* pHeader()const{ return &k5hdr;}
  //  void Reset_header(){;} ///< Reset sequential number of header
  /// Set Framesize (bytes)
  void Set_Framesize(const uint32_t sz){;}
  /// Set Headr and DAS parameter to the packet.
  void Set_Head_par(const time_t t1, DAS_Info* dasinf=0);
  /// Set Das Info
  const int Set_DasInfo(const DAS_Info& ds);
};

inline  void K5VSSP32_Frame::Set_Head_par(const time_t t1, DAS_Info* dasinf){
    VLBI_Frame::Set_Head_par(t1,dasinf);
    if(dasinf != 0){
	k5hdr.Set_Nch(dasinf->Nch());
	k5hdr.Set_Qbit(dasinf->Qbit());
	k5hdr.Set_Srate(dasinf->Srate()); // input in Hz
    }
}

inline int K5VSSP32_Frame::Make_frame( uint8_t* pkt,  const uint8_t* src,
				       const int  data_size){
    Set_header(pkt);
    pkt += K5VSSP32_Hdr_size;
    memcpy( pkt, src, data_size );
    return data_size + K5VSSP32_Hdr_size;
}

//========== K5VSSP ===to   Here  ====================

//========== K5VSI ===from Here  ====================

class K5VSI_Frame: public VLBI_Frame{
    //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
    //  K5VSSP32_Hdr k5hdr;  ///<K5VSSP32 Hearder
    string stname;
    //  string stid;
public:
    /// Default constructor
    K5VSI_Frame(){;}
    /// Returns Frame type
    const VLBI_Frame_t Type()const{return K5VSI_frm;}
    string TypeName(){return string("K5VSI");}
    //  int Decompose_frame(const uint8_t* buf, uint32_t size ){return 0;}
    int Make_frame( uint8_t* buf, const uint8_t* src, const int size );
    /// Headersize (bytes)
    int Headsize()const{return 0;}
    //  int Set_Sparam(const int n, const int q, const uint32_t s);
    //  void Set_Epoch(const Time_dat& t1);
    //  void Set_Epoch(const time_t t1);
    //  int Set_header(const uint8_t* buf, const int i=-1){return 0;}
    //  void Set_StnameID(const char* name, const char* id=0){ stname = name; }
    //  void Reset_header(){;} ///< Reset sequential number of header
    //  const K5VSI32_Hdr* pHeader()const{ return &k5hdr;}
    void Set_Framesize(const uint32_t sz){;}
};

inline int K5VSI_Frame::Make_frame( uint8_t* pkt, const uint8_t* src,
				    const int data_size){
    //  Set_header(pkt);
    //  pkt += K5VSSP32_Hdr_size;
    memcpy( pkt, src, data_size );
    return data_size;
}

//========== K5VSI ===to   Here  ====================
#endif
